package F5::MBIP::Sidecar;

use strict;
use warnings;

use Net::NATS::Client;
use YAML::XS;
use File::Spec;
use File::Find;
use File::Basename;
use File::Slurp;
use File::Copy;
use File::Path;
use MIME::Base64;
use F5::ProtocolBuffers;
use JSON;
use POSIX;
use Digest::MD5;

use F5::DateTimeUtils;
use F5::Deploy;
use F5::ASMConfig::GuaranteedClient;
use F5::ConfigSet;
use F5::SetActive;
use F5::Db::Single;
use F5::Log::Job;
use F5::DeviceConfig;
use F5::SystemUtils;
use F5::Captcha;

our $are_protobufs_loaded = 0;

our $icb_heartbeat_message_type = 'declTmm::dpcMessageV2';

our %HANDLE_ELEMENTS = map { ($_ => 1) }
  qw(waf_policy
     waf_remote_log_profile
     application_security);

our %LOG_PROFILE_FILTER = (0 => 'illegal',
                           1 => 'illegal-including-staged-signatures',
                           2 => 'all',
                          );

our %EXCEPTION_MESSAGE_REASON = (
    0 => 'unspecified',
    1 => 'oversized',
    2 => 'delete',
);

#----------------------------------------------------------------------

sub new {
    my($class, %params) = @_;
    my $self = bless {}, $class;

    $self->init(%params);
    return $self;
}

#----------------------------------------------------------------------

sub init {
    my($self, %params) = @_;

    my $config = YAML::XS::LoadFile($params{config_file});
    for my $k (keys %$config) {
        $self->{$k} = $config->{$k};
    }

    for my $k (keys %params) {
        $self->{$k} = $params{$k};
    }

    for my $k (keys %ENV) {
        if ($k =~ /^F5WAF_MBIP_SIDECAR_(.*)$/) {
            $self->{lc $1} = $ENV{$k};
        }
    }

    $self->{asmconfig_client} = F5::ASMConfig::GuaranteedClient::get_client(username => 'mbip_sidecar');
    F5::Db::Single->init;
    $self->{dbh} = F5::Db::Single->instance();
    $self->{config_set} = F5::ConfigSet->new(dbh => $self->{dbh},
                                             logger => $self->{logger},
                                             client => $self->{asmconfig_client});

    $self->{messages_handled_count} = 0;
    $self->{log_profiles_by_name} = {};
    $self->{log_profiles_by_id} = {};
    $self->{application_security_objs} = {};

    my $device_config = F5::DeviceConfig->new(dbh => $self->{dbh}, logger => $self->{logger});
    $device_config->apply();

    $self->{captcha} = F5::Captcha->new(dbh => $self->{dbh},
                                        logger => $self->{logger},
                                        asmconfig_client => $self->{asmconfig_client});
    $self->{captcha}->generate_captchas(generate_count => 1, out_dir_suffix => "_init", check_if_captcha_in_use => 1); # generate a single captcha on startup - blocking
    $self->{is_captcha_in_use} = 0;

    if ($self->{initial_full_dir}) {
        $self->{initial_full_dir} = F5::Deploy::map_path($self->{initial_full_dir});
        $self->{source_full_dir} = F5::Deploy::map_path($self->{source_full_dir});
        $self->create_initial_full_dir;
    }

    $self->{generation} = 0;
    $self->{sequence} = 0;
}

#----------------------------------------------------------------------

sub reinit {
    my($self, %params) = @_;

    my $rv_proto = $self->load_protobuf_definitions();

    my $rv_nats = 1;
    if ($rv_proto && !$self->{nats_client}) {
        $rv_nats = $self->connect_to_nats();
    }

    return $rv_proto && $rv_nats;
}

#----------------------------------------------------------------------

sub connect_to_nats {
    my($self, %params) = @_;

    $self->{logger}->debug("Connecting to NATS") if $self->{logger};
    my $username_prefix = '';
    if ($self->{nats_username}) {
        $username_prefix = $self->{nats_username};
        if ($self->{nats_password}) {
            $username_prefix .= ":$self->{nats_password}";
        }
        $username_prefix .= '@';
    }

    my $connect_rv;
    eval {
        # CPAN module Net::NATS::Client uses module URI under the hood in order to parse the URI. The
        # username and password cannot be extracted if an unknown protocol is used. Thus, although the
        # actual protocol is NATS internal protocol, this is listed as "ftp" in order for URI parsing to
        # work correctly.
        my %socket_args;
        if ($self->{nats_connect_timeout}) {
            $socket_args{Timeout} = $self->{nats_connect_timeout};
        }
        if ($self->{ssl_ca_file}) {
            $socket_args{SSL_ca_file} = $self->{ssl_ca_file};
        }
        if ($self->{ssl_cert_file}) {
            $socket_args{SSL_cert_file} = $self->{ssl_cert_file};
        }
        if ($self->{ssl_key_file}) {
            $socket_args{SSL_key_file} = $self->{ssl_key_file};
        }
        if ($self->{ssl_version}) {
            $socket_args{SSL_version} = $self->{ssl_version};
        }
        if ($self->{ssl_cipher_list}) {
            $socket_args{SSL_cipher_list} = $self->{ssl_cipher_list};
        }
        $self->{nats_client} = Net::NATS::Client->new(uri => 'ftp://' . "$username_prefix$self->{nats_host}:$self->{nats_port}",
                                                      socket_args => \%socket_args,
                                                     );

        $connect_rv = $self->{nats_client}->connect();
    };
    if ($@ || ! $connect_rv) {
        my $error_msg = $@ || '';
        $error_msg =~ s| at /\S+ line \d+$||;
        if ($error_msg) {
            $error_msg = "($error_msg)";
        }
        $self->{logger}->error("Couldn't connect to NATS $error_msg") if $self->{logger};
        delete $self->{nats_client};
        return;
    }
    $self->{logger}->info("Connected to NATS") if $self->{logger};

    my $subscription = $self->{nats_client}->subscribe($self->{nats_subject}, sub {
                                                           my ($message) = @_;
                                                           eval {
                                                               $self->handle_message(message => $message->data);
                                                           };
                                                           # this error is likely to be an "ASMConfigException" object
                                                           if (my $err = $@) {
                                                               $self->{logger}->error("Failed to handle NATS message:") if $self->{logger};
                                                               $self->{logger}->error($err) if $self->{logger};
                                                           }
                                                  });

    return 1;
}

#----------------------------------------------------------------------
# Protobuf defintions should only be loaded once for a process. Once .proto files from a particular
# path have been loaded, this will be intentionally skipped.

sub load_protobuf_definitions {
    my($self) = @_;

    return 1 if ($are_protobufs_loaded);

    if (! -d $self->{protobuf_file_path}) {
        $self->{logger}->info("Couldn't find protobuf defintions") if $self->{logger};
        return $are_protobufs_loaded;
    }

    $self->{logger}->debug("Loading protobuf defintions") if $self->{logger};

    $are_protobufs_loaded = F5::ProtocolBuffers::init(base_dir => $self->{protobuf_file_path},
                                                      package => $self->{protobuf_base_package},
                                                      imports_dir => $self->{protobuf_imports_dir},
                                                     );

    if ($are_protobufs_loaded) {
        $self->{logger}->info("Loaded protobuf defintions") if $self->{logger};
    }

    return $are_protobufs_loaded;
}

#----------------------------------------------------------------------

sub send_icb_heartbeat {
    my($self) = @_;

    if (! $are_protobufs_loaded) {
        $self->{logger}->debug("Skipping ICB heartbeat (protobuf definitions are not loaded)") if $self->{logger};
        return;
    }

    if (! $self->{nats_client}) {
        $self->{logger}->debug("Skipping ICB heartbeat (NATS Client not connected)") if $self->{logger};
        return;
    }

    my $message = $icb_heartbeat_message_type->new();
    # default accessors names are: 'set_' + protobuf message field name
    $message->set_generation($self->{generation});
    $message->set_seq_start($self->{sequence});
    $self->{logger}->trace("Created $icb_heartbeat_message_type message:") if $self->{logger};
    $self->{logger}->trace($message) if $self->{logger};

    my $bytes = $icb_heartbeat_message_type->encode($message);

    if (!$self->{generation} && !$self->{sequence}) {
        $self->{logger}->info("Sending initial heartbeat message") if $self->{logger};
    }

    $self->{nats_client}->request($self->{heartbeat_subject}, $bytes , sub {
                                      my ($reply) = @_;
                                      if (defined $reply && $self->{logger}) {
                                          $self->{logger}->trace("ICB reply:");
                                          $self->{logger}->trace($reply);
                                      }
    });
    $self->{logger}->debug("Sent heartbeat message") if $self->{logger};
}

#----------------------------------------------------------------------

sub run {
    my($self) = @_;

    my $last_pulse_time = time;
    my $last_captcha_time = 0;
    my $last_icb_heartbeat_time = 0;

    while (1) {
        $self->reinit();

        my @potential_poll_times;

        if ($self->{icb_heartbeat_interval}) {
            push(@potential_poll_times,
                 F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $self->{icb_heartbeat_interval},
                                                                    start_time => $last_icb_heartbeat_time));
        }
        if ($self->{pulse_msg_interval}) {
            push(@potential_poll_times,
                 F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $self->{pulse_msg_interval},
                                                                    start_time => $last_pulse_time));
        }

        my $poll_time = List::Util::min(grep { $_ } (@potential_poll_times, 5));
        if ($self->{nats_client}) {
            eval {
                $self->{logger}->trace("Going to poll for $poll_time seconds") if $self->{logger};
                $self->{nats_client}->wait_for_op($poll_time);
            };
            if ($@) {
                my $error_msg = $@ || '';
                $error_msg =~ s| at /\S+ line \d+$||;
                if ($error_msg) {
                    $error_msg = "($error_msg)";
                }
                $self->{logger}->error("Couldn't poll NATS server $error_msg") if $self->{logger};
                delete $self->{nats_client};
            }
        }
        else {
            sleep $poll_time;
        }

        if ($self->{pulse_msg_interval} &&
            F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $self->{pulse_msg_interval},
                                                               start_time => $last_pulse_time) == 0) {
            $self->log_pulse_message;
            $last_pulse_time = time;
        }
        if ($self->{captcha_generate_interval} && $self->{is_captcha_in_use} &&
            F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $self->{captcha_generate_interval},
                                                               start_time => $last_captcha_time) == 0) {
            $self->spawn_generate_captchas;
            $last_captcha_time = time;
        }
        if ($self->{icb_heartbeat_interval} &&
            F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $self->{icb_heartbeat_interval},
                                                               start_time => $last_icb_heartbeat_time) == 0) {

            $self->send_icb_heartbeat();
            $last_icb_heartbeat_time = time;
        }

        if ($self->{current_captcha_pid}) {
            $self->reap_spawned_processes();
        }
    }
}

#----------------------------------------------------------------------

sub log_pulse_message {
    my($self) = @_;

    $self->{logger}->info("$self->{messages_handled_count} messages handled since startup") if $self->{logger};
}

#----------------------------------------------------------------------

sub handle_message {
    my($self, %params) = @_;

    my $reason = $params{reason};
    my $message = $params{message};

    my $proto_struct = $params{proto_struct} || $self->{protobuf_base_struct};
    my $parent_proto_struct = $params{parent_proto_struct} || '';

    my $proto_struct_decoder = $self->{protobuf_base_package} . '::' . $proto_struct;
    my $message_rec;
    eval {
        $message_rec = $proto_struct_decoder->decode($message);
    };
    if (my $err = $@) {
        $self->{logger}->error("Couldn't decode $proto_struct_decoder message. Error: $err") if $self->{logger};
        return;
    }

    if (! $message_rec) {
        return;
    }

    if (ref($message_rec) eq 'declTmm::unsatisfiableMsgV2' &&
        $message_rec->{generation} == $self->{generation} &&
        $message_rec->{seq_start} == $self->{sequence}) {

        $self->{generation} = 0;
        $self->{sequence} = 0;
        return;
    }

    if (ref($message_rec) eq 'declTmm::exceptionObject' && exists $message_rec->{reason}) {
        $reason = $EXCEPTION_MESSAGE_REASON{$message_rec->{reason}};
    }

    if (ref($message_rec) eq 'declTmm::createMsgV2' && exists $message_rec->{header} &&
        ref($message_rec->{header}) eq 'declTmm::headerV2' && $message_rec->{header}->{end_flag} == 1) {
        $self->{generation} = $message_rec->{header}->{generation} || 0;
        $self->{sequence} = $message_rec->{header}->{seq_end} || 0;
    }

    if ($message_rec->{embedded} && ref($message_rec->{embedded}) ne 'ARRAY' && $message_rec->{embedded}->{type_url} =~ /declTmm\.(.*)/) {
        my $embedded_proto_type = $1;
        $self->{logger}->info("Handling embedded $message_rec->{embedded}->{type_url} message") if $self->{logger};
        $self->handle_message(reason => $reason,
                              message => $message_rec->{embedded}->{value},
                              proto_struct => $embedded_proto_type,
                              parent_proto_struct => $proto_struct,
                             );
    }
    elsif ($message_rec->{embedded} && ref($message_rec->{embedded}) eq 'ARRAY') {
        for my $embedded_message (@{$message_rec->{embedded}}) {
            if ($embedded_message->{type_url} =~ /declTmm\.(.*)/) {
                my $embedded_proto_type = $1;
                $self->{logger}->info("Handling embedded $embedded_message->{type_url} message") if $self->{logger};
                $self->handle_message(reason => $reason,
                                      message => $embedded_message->{value},
                                      proto_struct => $embedded_proto_type,
                                      parent_proto_struct => $proto_struct,
                                     );
            }
        }
    }
    elsif ($HANDLE_ELEMENTS{$proto_struct}) {
        my $element_name = $proto_struct;

        my $function_name;
        if ($parent_proto_struct eq 'createMsgV2') {
            $function_name = 'add_or_update_';
            $self->{logger}->debug("Add/update for $proto_struct named $message_rec->{name}") if $self->{logger};
        }
        elsif ($parent_proto_struct eq 'exceptionObject') {
            if ($reason eq 'delete') {
                $function_name = 'delete_';
                $self->{logger}->info("Delete for $proto_struct id $message_rec->{id}") if $self->{logger};
                #   in this case the AsyncAPI object `$message_rec` contains only the AsyncAPI UUID (not the ASMConfig UUID):
                #     declTmm::waf_policy { id: 'some-AsyncAPI-uuid' }
                #   rather than a full blown object:
                #     declTmm::waf_policy { name:'...', id:'...', filename:'...' }
            }
            if ($reason eq 'unspecified') {
                $self->{logger}->error("Unspecified exception while handling $proto_struct") if $self->{logger};
                return;
            }
            if ($reason eq 'oversized') {
                $self->{logger}->error("Oversized exception while handling $proto_struct") if $self->{logger};
                return;
            }
        }
        else {
            $self->{logger}->error("Unrecognized op struct $parent_proto_struct while handling element $proto_struct") if $self->{logger};
            return;
        }
        $function_name .= $element_name;

        my %job_attributes = (logger => $self->{logger},
                              operationId => $function_name,
                              component => 'mbip_sidecar',
                             );

        if ($message_rec->{name}) {
            my $rest_element_name = $element_name;
            $rest_element_name =~ s/_(.)/\U$1/g;
            $job_attributes{$rest_element_name . "Name"} = $message_rec->{name};
        }

        my $import_content;
        if ($parent_proto_struct eq 'createMsgV2') {
            if ($message_rec->{filename}) {
                if (! -f $message_rec->{filename} || -z $message_rec->{filename}) {
                    $self->{logger}->error("No such policy file or empty file at: $message_rec->{filename}") if $self->{logger};
                    return;
                }
                $import_content = File::Slurp::read_file($message_rec->{filename});
            }
            elsif ($message_rec->{importContent}) {
                $import_content = delete $message_rec->{importContent};
            }
            else {
                $self->{logger}->error("Either filename or import content must be provided") if $self->{logger};
            }

            $job_attributes{policySize} = length($import_content);
            $job_attributes{policyChecksum} = Digest::MD5->new->add($import_content)->hexdigest;

            if ($self->{logger} && $self->{logger}->is_trace()) {
                $message_rec->{importContent} = substr($import_content, 0, 128);
                $message_rec->{importContent} .= "... (truncated to 128 characters for logging)";
                $self->{logger}->trace($message_rec);
            }

            $message_rec->{importContent} = $import_content;
        }

        my $log_message = "Received $proto_struct message";
        $log_message = $log_message . " (reason $reason)" if defined $reason;
        $self->{logger}->debug($log_message) if $self->{logger};

        $self->{logger}->debug("Calling function_name $function_name");
        $self->{logger}->debug($message_rec);

        my $job = F5::Log::Job->new(%job_attributes);
        $job->set_started();
        $self->$function_name(%$message_rec, job => $job);
        $job->set_completed();
    }
    else {
        if ($self->{logger} && $self->{logger}->is_trace()) {
            my $json = JSON->new->allow_nonref->convert_blessed;;
            my $pretty_json = $json->pretty->encode($message_rec);
            my $msg = "$proto_struct_decoder message";
            $msg = $msg . " (reason $reason)" if defined $reason;
            $self->{logger}->trace("Handling $msg: $pretty_json");
        }
    }
}

#----------------------------------------------------------------------

sub delete_waf_policy {
    my($self, %params) = @_;
    my $api_uuid = $params{id};

    my $policy_name = $self->{api_id_policy_name_map}->{$api_uuid};

    $self->{logger}->info("Deleting policy $policy_name (api_uuid: $api_uuid)") if $self->{logger};

    my $call_flags = F5::ASMConfig::CallFlags->new();

    my $policy_criteria = F5::ASMConfig::Policy->new(policy_name => $policy_name);

    $self->{asmconfig_client}->delete_policy($self->{asmconfig_client}->{session},
                                             $call_flags,
                                             $policy_criteria);
}

#----------------------------------------------------------------------

sub add_or_update_waf_policy {
    my($self, %params) = @_;
    my $api_uuid = $params{id};

    my $import_policy_name = $params{name};
    my $filename = $params{filename};

    $self->{api_id_policy_name_map}->{$api_uuid} = $import_policy_name;

    $self->{logger}->info("Applying updated policy $import_policy_name (api_uuid: $api_uuid)") if $self->{logger};

    my $call_flags = F5::ASMConfig::CallFlags->new();

    if (! $filename) {
        my $fh;
        ($fh, $filename) = File::Temp::tempfile();
        print $fh $params{importContent};
        close $fh;
    }

    $self->{logger}->info("Preparing importing for policy $import_policy_name")  if $self->{logger};
    my $import_export_properties = F5::ASMConfig::ImportExportProperties->new(filename => $filename,
                                                                              policy_name => $import_policy_name,
                                                                              force_policy_name_flg => 1,
                                                                              overwrite => 1,
                                                                             );

    my $import_rv = $self->{asmconfig_client}->import_policy($self->{asmconfig_client}->{session},
                                                $call_flags,
                                                $import_export_properties);

    unlink $filename;

    my $set_active_rv = F5::SetActive::set_active(dbh => $self->{dbh},
                                                  logger => $self->{logger},
                                                  job => $params{job},
                                                  is_daemon => 0,
                                                  use_daemon => 0,
                                                  policy_id => $import_rv->{policy}->{policy_id},
                                                  update_ramcache => 0,
                                                  do_update_core => 1,
                                                  expanded_return_value => 1,
                                                  do_save_policy_history => 0,
                                                  asm_config_handler => $self->{asmconfig_client});
    if (!$set_active_rv->{completed_successfully}) {
        my $msg = "Failed to compile Policy $import_policy_name";
        $self->{logger}->error($msg)  if $self->{logger};
        die {"error_message" => $msg};
    }

    if (! $self->{is_captcha_in_use}) {
        $self->{is_captcha_in_use} = $self->{captcha}->is_captcha_in_use;
    }
}

#----------------------------------------------------------------------

sub add_or_update_waf_remote_log_profile {
    my($self, %params) = @_;

    my $log_profile = \%params;
    $self->{logger}->info("Adding logging profile $log_profile->{name} ($log_profile->{id})") if $self->{logger};

    my @ip_address_segments = unpack('C*', pack('Q', $log_profile->{addr}->{a2}));
    $log_profile->{ip_address} = join('.', @ip_address_segments[4..7]);
    $self->{log_profiles_by_id}->{$log_profile->{id}} = $log_profile;
    $self->{log_profiles_by_name}->{$log_profile->{name}} = $log_profile;

    $self->prepare_logging_profiles(%params);
}

#----------------------------------------------------------------------

sub delete_waf_remote_log_profile {
    my($self, %params) = @_;

    $self->{logger}->info("Deleting logging profile $params{name}") if $self->{logger};
    my $log_profile = $self->{log_profiles_by_name}->{$params{name}};
    delete $self->{log_profiles_by_name}->{$log_profile->{name}};
    delete $self->{log_profiles_by_id}->{$log_profile->{id}};

    $self->prepare_logging_profiles();
}

#----------------------------------------------------------------------

sub add_or_update_application_security {
    my($self, %params) = @_;

    my $waf_log_profile_id = $params{wafLogProfileId} || '';
    $self->{logger}->info("Adding application security $params{name} (virtualServerId = $params{virtualServerId}, wafLogProfileId = $waf_log_profile_id)") if $self->{logger};
    $self->{application_security_objs}->{$params{name}} = { %params };

    $self->prepare_logging_profiles(%params);
}

#----------------------------------------------------------------------

sub delete_application_security {
    my($self, %params) = @_;

    $self->{logger}->info("Deleting application security $params{name}") if $self->{logger};
    delete $self->{application_security_objs}->{$params{name}};

    $self->prepare_logging_profiles(%params);
}

#----------------------------------------------------------------------

sub prepare_logging_profiles {
    my($self, %params) = @_;

    $self->{logger}->debug("Updating logging profiles") if $self->{logger};

    my @vs_configs;
    for my $security_obj (values %{$self->{application_security_objs}}) {
        if (! $security_obj->{wafLogProfileId}) {
            $self->{logger}->info("WAF Security $security_obj->{name} has no logging profile -- skipping") if $self->{logger};
            next;
        }

        my $log_profile = $self->{log_profiles_by_id}->{$security_obj->{wafLogProfileId}};
        if (! $log_profile) {
            $self->{logger}->info("WAF Log Profile $security_obj->{wafLogProfileId} doesn't exist -- skipping") if $self->{logger};
            next;
        }
        my $logging_config = { logging => {
                                           dest_ip => $log_profile->{ip_address},
                                           dest_port => $log_profile->{port},
                                           filter => $LOG_PROFILE_FILTER{$log_profile->{filterRequestType}},
                                           content => {
                                                       max_message_size => '64k',
                                                      },
                                          },
                             };

        my $vs_config = $self->{config_set}->build_logging_config(policy => { vs_info => { $security_obj->{virtualServerId} => $logging_config } });
        push(@vs_configs, @$vs_config);
    }

    $self->{config_set}->prepare_all_logging_profiles(vs_config => \@vs_configs);
    my $set_active_rv = F5::SetActive::set_active(dbh => $self->{dbh},
                                                  logger => $self->{logger},
                                                  job => $params{job},
                                                  do_update_core => 1,
                                                  asm_config_handler => $self->{asmconfig_client});
}

#----------------------------------------------------------------------

sub create_initial_full_dir {
    my($self) = @_;

    my $set_active_rv = F5::SetActive::set_active(dbh => $self->{dbh},
                                                  logger => $self->{logger},
                                                  do_update_core => 1,
                                                  asm_config_handler => $self->{asmconfig_client});

    mkpath($self->{initial_full_dir});
    find({
          no_chdir => 1,
          follow_fast => 1,
          wanted => sub {
              my $filename = basename($File::Find::name);
              copy($File::Find::name, "$self->{initial_full_dir}/$filename");
          },
         },
         $self->{source_full_dir});
}

#----------------------------------------------------------------------

sub spawn_generate_captchas {
    my($self, %params) = @_;

    if ($self->{current_captcha_pid} = fork) {
        # in parent
        $self->{logger}->debug("Forked to the parent with child pid $self->{current_captcha_pid}") if $self->{logger};
    }
    else {
        # in child
        exit $self->{captcha}->generate_captchas(%params) || 0;
    }
}

#----------------------------------------------------------------------

sub reap_spawned_processes {
    my($self, %params) = @_;

    if ($self->{current_captcha_pid}) {
        my $rv = waitpid($self->{current_captcha_pid}, WNOHANG);
        if ($rv) {
            my $exit_code = $? >> 8;
            $self->{logger}->debug("Generate captchas: child exited with code: $exit_code") if $self->{logger};
            delete $self->{current_captcha_pid};
        }
    }
}

#----------------------------------------------------------------------

1;
